#include <stdio.h>
#include <math.h>
#include <vector>
#include <map>

struct Tank			
{
	char name[64], nation[4], type[4];
	int tier, mmrange, id, common;		// common => used for renormalization
	double minxp, xpscale;
};

std::vector<Tank> g_tanks;
std::map<int, int> g_tankmap;		// maps tank id to index in g_tanks

int LoadTanksTest()
{
	FILE *pFile = fopen("test_tanklist.csv", "rt");
	if(!pFile) {
		printf("Tank list file not found\n");
		return 0;		
	}

	// skip first row because it's just headers
	char pBuf[1024]; char *r = fgets(pBuf, 1024, pFile);
	while(r)
	{
		r = fgets(pBuf, 1024, pFile);			 
		if (!r) break;							// fgets still completes if EOF found halfway through

		Tank t;
		// format:name,id,tier,nat,type, mm x6
		int c = sscanf_s(pBuf, "%[^,],%i,%i,%[^,],%[^,], %i,%*lf,%*[^,],%lf,%lf,%i\n",
			t.name, 64, &t.id, &t.tier, t.nation, 4, t.type, 4,
			&t.mmrange, &t.minxp, &t.xpscale, &t.common);
		if (!c) break;

		g_tanks.push_back(t);
		g_tankmap.insert(std::pair<int,int>(t.id, g_tanks.size()-1));
	}
	return 1;
}

struct TankWeekData
{
	int id, battles, wins, damdealt, frags, spots, def, cap, xp, startxp;
};

struct PlayerData
{
	unsigned int accid;
	int battles, startbat;
	std::vector<TankWeekData> tanks;
};


int ReadAccountTest(FILE *pFile, PlayerData &p)
{
	if (!fread(&p.accid, 4, 1, pFile)) return 0;
	fread(&p.battles, 4, 1, pFile);
	fread(&p.startbat, 4, 1, pFile);
	int nt; fread(&nt, 4, 1, pFile);
	p.tanks.resize(nt);
	for (int t=0; t<nt; t++)
	{
		fread(&p.tanks[t].id, 4, 1, pFile);
		fread(&p.tanks[t].battles, 4, 1, pFile);
		fread(&p.tanks[t].wins, 4, 1, pFile);
		fread(&p.tanks[t].damdealt, 4, 1, pFile);
		fread(&p.tanks[t].frags, 4, 1, pFile);
		fread(&p.tanks[t].spots, 4, 1, pFile);
		fread(&p.tanks[t].def, 4, 1, pFile);
		fread(&p.tanks[t].cap, 4, 1, pFile);
		fread(&p.tanks[t].xp, 4, 1, pFile);
		fread(&p.tanks[t].startxp, 4, 1, pFile);
	}
	return 1;
}


struct TierAvgExp { double win, dmg, frag, spot, def, cap, weight; };
static TierAvgExp g_tierAvg[10] = {		// from 150816 EU avgs exc scout/arty
	{ 0.477, 88.9, 0.68, 0.90, 0.53, 1.0, 0.40 },
	{ 0.490, 118.2, 0.66, 0.85, 0.65, 1.0, 0.41 },
	{ 0.495, 145.1, 0.59, 1.05, 0.51, 1.0, 0.44 },
	{ 0.492, 214.0, 0.60, 0.81, 0.55, 1.0, 0.44 },
	{ 0.495, 388.3, 0.75, 0.93, 0.63, 1.0, 0.60 },
	{ 0.497, 578.7, 0.74, 0.93, 0.52, 1.0, 0.70 },
	{ 0.498, 791.1, 0.76, 0.87, 0.58, 1.0, 0.82 },
	{ 0.497, 1098.7, 0.79, 0.87, 0.58, 1.0, 1.00 },
	{ 0.498, 1443.2, 0.86, 0.94, 0.56, 1.0, 1.23 },
	{ 0.498, 1963.8, 1.04, 1.08, 0.61, 1.0, 1.60 }};

double CalcWN9Base (double dmg, double frag, double spot, double def, Tank &ev)
{
	TierAvgExp &avg = g_tierAvg[ev.mmrange == 3 ? ev.tier : ev.tier-1];
	double rdmg = dmg / avg.dmg; double rfrag = frag / avg.frag;
	double rspot = spot / avg.spot;	double rdef = def / avg.def;
	return 0.7*rdmg + 0.25*sqrt(rfrag*rspot) + 0.05*sqrt(rfrag*sqrt(rdef));
}


struct XYW { double x, y, w; };

double WeightedLSR (std::vector<XYW> &data, double *inter, double *slope)
{
	// find weighted means
	double sumwx=0, sumwy=0, sumw=0;
	for (std::vector<XYW>::iterator it=data.begin(); it!=data.end(); ++it)
	{
		sumwx += (*it).w * (*it).x;
		sumwy += (*it).w * (*it).y;
		sumw += (*it).w;
	}
	double wxbar = sumwx / sumw;
	double wybar = sumwy / sumw;

	double sumxy=0, sumx2=0, sumy2=0;
	for (std::vector<XYW>::iterator it=data.begin(); it!=data.end(); ++it)
	{
		double xsubxbar = (*it).x - wxbar;
		double ysubybar = (*it).y - wybar;
		sumxy += (*it).w * xsubxbar * ysubybar;
		sumx2 += (*it).w * xsubxbar * xsubxbar;
		sumy2 += (*it).w * ysubybar * ysubybar;
	}
	*slope = sumxy / sumx2;
	*inter = wybar - *slope*wxbar;
	return sumxy / sqrt(sumx2*sumy2);
}


struct RParam { double inter, slope, tankstat, accstat, adjstat; };
struct TankStat { int index, battles; double stat; };

// stat needs to be *average*, not total
// no filtering: input data should be prefiltered for:
// - minimum battle count, zero battle tanks, max in one tank etc
// input: list of tankindex + parameter + battle count 
// input & output: list of regression slope + inter

void ParamRegression(const std::vector<std::vector<TankStat>> &players,
	RParam rparam[], bool regress)
{
	std::vector<XYW> tankdata[500];
	std::vector<double> skilltot(500), tanktot(500);

	for (size_t p=0; p<players.size(); p++)
	{
		const std::vector<TankStat> &tanks = players[p];

		// generate player skill from tank stats
		int bc=0; double skill=0.0;
		for (size_t t=0; t<tanks.size(); t++)
		{
			const TankStat &tank = tanks[t];
			RParam &rp = rparam[tank.index];
			bc += tank.battles;
			skill += (tank.stat - rp.inter) * tank.battles / rp.slope;
		}
		skill /= (double)bc;

		// now add tanks as data points
		for (size_t t=0; t<tanks.size(); t++)
		{
			const TankStat &tank = tanks[t];
			XYW xyw = { skill, tank.stat, (double)tank.battles };
			tankdata[tank.index].push_back(xyw);
			skilltot[tank.index] += skill * tank.battles;
			tanktot[tank.index] += tank.stat * tank.battles;
		}
	}

	// generate new slope parameters
	for (size_t t=0; t<g_tanks.size(); t++)
	{
		if (tankdata[t].size() <= 10) continue;
		if (!regress) {
			rparam[t].inter = 0;
			rparam[t].slope = tanktot[t] / skilltot[t];
		}
		else WeightedLSR(tankdata[t], &rparam[t].inter, &rparam[t].slope);
	}
}

void RenormInterSlope(RParam rparam[])
{
	double totalslope=0.0, totalinter=0.0, totalbattles=0.0;
	for (size_t t=0; t<g_tanks.size(); t++)
	{
		if (!g_tanks[t].common || g_tanks[t].tier != 10) continue;		// only renormalize with common 10 tanks
		totalinter += rparam[t].inter;
		totalslope += rparam[t].slope;
		totalbattles += 1;
	}
	totalinter /= totalbattles;
	totalslope /= totalbattles;
	printf ("pre-norm t10 inter = %.6f, t10 slope = %.6f\n", totalinter, totalslope);

	for (size_t t=0; t<g_tanks.size(); t++)
	{
		// renormalize tanks to 0 inter, 1 slope
		rparam[t].inter -= totalinter;
		rparam[t].slope /= totalslope;
	}
}


#define SRVDATE "eu_150830_151108"

int main()
{
	LoadTanksTest();

	std::vector<std::vector<TankStat>> playerstats;
	std::vector<int> tankcount(500), tankbat(500);

	// build TankStat arrays from diff file
	FILE *pFile = fopen("test_data_" SRVDATE ".dat","rb");
	if (!pFile) { printf("Test data file not found\n"); return 0; }
	int c=0; while(1)
	{
		PlayerData p;
		if (!ReadAccountTest(pFile, p)) break;
		if (!(++c % 10000)) printf ("%i accounts processed\n", c);
		if (p.battles <= 0 || p.battles > 4000) continue;
		if (p.startbat < 2500) continue;

		// first scale weekly results by earned XP and merge them
		struct TankColl { double win, dmg, frag, spot, def; int battles; } tankcoll[500];
		memset(tankcoll, 0, 500*sizeof(TankColl));
		for (size_t i=0; i<p.tanks.size(); i++)
		{
			TankWeekData &t = p.tanks[i];
			std::map<int,int>::iterator it = g_tankmap.find(t.id);
			if (it == g_tankmap.end()) {
				printf("Tank ID %i not found\n", t.id);
				continue;
			}
			int tidx = it->second;
			Tank &ev = g_tanks[tidx];

			double xpmid = t.startxp + (t.xp>>1);
			if (xpmid < ev.minxp) continue;
			if (xpmid > 400000) xpmid = 400000;				// cap the worst cases
			if (xpmid < 5000) xpmid = 5000;					// cap the worst cases

			const double log3i = 1.0 / log(3.0);
			double xpoffset = -ev.xpscale * ((log(xpmid*0.001) * log3i) - 4);

			tankcoll[tidx].battles += t.battles;
			tankcoll[tidx].win += t.wins * (1 + xpoffset*1.5/10.0); // 10% damage worth ~1.5% winrate
			tankcoll[tidx].dmg += t.damdealt * (1 + xpoffset);
			tankcoll[tidx].frag += t.frags * (1 + xpoffset);
			tankcoll[tidx].spot += t.spots * (1 + xpoffset);
			tankcoll[tidx].def += t.def * (1 + xpoffset);			// maybe should be squared...
		}

		// now then go through tanksum amd create the TankStat vectors
		std::vector<TankStat> tankstats;
		for (size_t i=0; i<g_tanks.size(); i++)
		{
			TankColl &t = tankcoll[i];
			if (t.battles < 5) continue;
			double idc = 1.0 / t.battles;
			tankcount[i]++; tankbat[i] += t.battles;

			double wn9base = CalcWN9Base (t.dmg*idc, t.frag*idc, t.spot*idc, t.def*idc, g_tanks[i]);
			TankStat twn9 = { i, t.battles, wn9base }; tankstats.push_back(twn9);
		}
		if (tankstats.size() > 1) playerstats.push_back(tankstats);
	}
	fclose (pFile);


	// Run the actual regression, normalizing after each step
	RParam rparam[500];
	for (int i=0; i<500; i++) { rparam[i].inter = 0.0; rparam[i].slope = 1.0; }

	printf ("Beginning regression...\n");
	for (int i=0; i<20; i++) {
		// note: other stats may need more regress=false steps
		ParamRegression(playerstats, rparam, i>2 ? true : false);
		RenormInterSlope(rparam);
	}


	// Write data to file
	FILE *pOut = fopen("exptanks_" SRVDATE "_test.csv","wt");
	fprintf (pOut, "name,id,tier,nation,type,players,battles,wn9exp,wn9scale\n");
	for (size_t t=0; t<g_tanks.size(); t++) 
	{
		const double t10avg = 1.0; 		// change for different params, eg. dmg = 1964
		double exp = rparam[t].inter + t10avg * rparam[t].slope;
		double scale = rparam[t].slope * t10avg / exp;

		// fudges here
		if (!strcmp(g_tanks[t].type, "SPG")) { scale *= 1.05; exp *= 1.1; } 
		else {
			// newbie MM -> pre newbie MM mapping
			if (g_tanks[t].tier == 1) exp *= 1.41;		// use 1.077 for winrate
			if (g_tanks[t].tier == 2) exp *= 1.15;		// use 1.044 for winrate
			if (g_tanks[t].tier == 3) exp *= 1.16;		// use 1.028 for winrate
		}

		fprintf(pOut, "%s,%i,%i,%s,%s,%i,%i,%.3f,%.3f\n", g_tanks[t].name, g_tanks[t].id, g_tanks[t].tier,
			g_tanks[t].nation, g_tanks[t].type, tankcount[t], tankbat[t], exp, scale);
	}
	fclose(pOut);

	return 0;
}

